/*
 * bootmenu - A program for spawning apps on a TTY
 * Developed especially for booting a graphical engine
 * but I guess it could be used for anything
 *
 * Copyright (C) 2003, Penguru Consulting, LLC
 * Written by:  Jordan Crouse (jordan@penguru.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/* Basic menu drawing */

/* This will draw your basic menu, without having to use ncurses.  Of 
   course, that means that this is Linux centric, and it probably won't 
   work on just any terminal.

   We go very simple here, not even counting on being able to 
   reverse the text on the selected item (so we draw an arrow instead).

   We are assuming, however, that you have arrow keys, and they are mapped
   for your device.  If not, then bootmenu might not be for you.
*/

/* The ever present TODO list:

 * Allow for a method of specifying a "up" key and a "down" key that
 may or may not be the arrow keys 
*/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "bootmenu.h"

#define ESC 1
#define EXT 2

static void print_items(struct app *apps, int hilite) {
  
  int x = 0;
  int y = 3;
  int i;

  for(i = 0; apps[i].title; i++) {
    printf("\033[%d;%dH", y, x); /* Move to the line */
    printf("\033[K"); /* Erase to the end of the line */
    
    if (i == hilite) 
      printf("--> ");
    else
      printf("    ");

    printf("%s\r\n", apps[i].title);
    y++;
  }
}

inline static unsigned char read_char(void) {
  unsigned char ch;

  if (read(0, &ch, 1) < 1) {
    LOG("basic.so Read error [%s]\n", strerror(errno));
    if (errno == EINTR || errno == EIO) {
      CLOSELOG();
      exit(0);
    }
  }

  return ch;
}

static struct app * basic_bootmenu(struct app *apps) {
  
  int val = 0;
  int count = 0;
  int mode = 0;

  for( ; apps[count].title; count++) ;

  printf("\033[H\033[J");  /* Clear the screen */
  printf("Bootmenu (Version 0.6)\r\n");

  print_items(apps, val);

  while(1) {
    unsigned char ch = read_char();

    if (mode == 0) {
      if (ch == 033) mode = ESC;
      if (ch == 10 || ch == 13) break; /* Enter */
    }   
    else if (mode == ESC) {
      if (ch == 'O' || ch == '[') mode = EXT;
      else mode  = 0;
    }
    else if (mode == EXT) {
      if (ch == 'A') {
	val = (val == 0) ? count - 1 : val - 1;
	print_items(apps, val);
      }

      if (ch == 'B') {
	val = (val == count - 1) ? 0 : val + 1;
	print_items(apps, val);
      }

      mode = 0;
    }
  }

  return &apps[val];
}

/* If we are building this as a plugin, then we provide
   the standard plugin hook  

   otherwise, we provide the global do_bootmenu() function
   called by spawn_bootmenu();
*/

#ifdef PLUGINS
struct app *plugin_bootmenu(struct app *apps) 
#else
struct app *do_bootmenu(struct app *apps) 
#endif

{
  return basic_bootmenu(apps);
}


	
    
      
      
  

  
