/*
 * bootmenu - A program for spawning apps on a TTY
 * Developed especially for booting a graphical engine
 * but I guess it could be used for anything
 *
 * Copyright (C) 2003, Penguru Consulting, LLC
 * Written by:  Jordan Crouse (jordan@penguru.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>  /* for offsetof */
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <termios.h>
#include <utmp.h>
#include <pwd.h>

#include <sys/wait.h>
#include <sys/ioctl.h>

#define _BOOTDIR "/etc/bootmenu"
#define _BOOTAPPSDIR "apps"  

#include "bootmenu.h"

/* List of ttys */
static struct tty ttylist[MAX_TTYS];

/* List of applications */
static struct app alist[MAX_APPS];

struct config bootmenu_config;

/* 
   The general procedure for correctly setting up the tty
   comes from getty.c from tinylogin (http://tinylogin.busybox.net/)

   Copyright (C) 1999,2000,2001 by Erik Andersen <andersee@codepoet.org>
*/

static int open_tty(const char *dev) {

  struct termio tp;

  /* Close the original files */

  close(1); 
  close(2);
  close(0);
  
  /* Open the tty (it should open on stdin) */

  if (open(dev, O_RDWR | O_NONBLOCK, 0) != 0) {
    LOG("Unable to open tty [%s]\n", dev);
    return -1;
  }

  if (dup(0) != 1) return -1;
  if (dup(0) != 2) return -1;

  /* Get the TERMIO settings */

  if (ioctl(0, TCGETA, &tp) < 0) 
    return -1;

  /* Flush the queues (probably not needed) */
  ioctl(0, TCFLSH, TCIOFLUSH);

  /* Force to 115200 baud.  Is that OK? */

  tp.c_cflag = CS8 | HUPCL | CREAD | B115200;
  tp.c_iflag = tp.c_lflag = tp.c_oflag = tp.c_line = 0;

  tp.c_cc[VMIN] = 1;
  tp.c_cc[VTIME] = 0;

  /* Set the TERMIO settings */
  ioctl(0, TCSETA, &tp);
  
  /* Block on input */
  fcntl(0, F_SETFL, fcntl(0, F_GETFL, 0) & ~O_NONBLOCK);
  
  /* Block on write too */
  fcntl(1, F_SETFL, fcntl(0, F_GETFL, 0) & ~O_NONBLOCK);

  return 0;
}

/* 
   This comes from init.c from busybox (http://busybox.net/)
   We do this to ensure that the terminal is set sanely for the
   application.  

   Copyright (C) 1999,2000,2001 by Erik Andersen <andersee@codepoet.org>

   FIXME:  Should we just set it up like this from the get go? 
*/

static void finalize_term(void) {

  struct termio tp;

  if (ioctl(0, TCGETA, &tp) < 0) return;

  /* Directly from BB init.  Thanks Erik */

  tp.c_cc[VINTR] = 3;	  /* C-c */
  tp.c_cc[VQUIT] = 28;	  /* C-\ */
  tp.c_cc[VERASE] = 127; /* C-? */
  tp.c_cc[VKILL] = 21;   /* C-u */
  tp.c_cc[VEOF] = 4;	  /* C-d */
  tp.c_cc[VSTART] = 17;  /* C-q */
  tp.c_cc[VSTOP] = 19;	  /* C-s */
  tp.c_cc[VSUSP] = 26;	  /* C-z */

  tp.c_line = 0;

  tp.c_cflag &= CBAUD | CBAUDEX | CSIZE | CSTOPB | PARENB | PARODD;
  tp.c_cflag |= CREAD | HUPCL | CLOCAL;
  
  tp.c_iflag = ICRNL | IXON | IXOFF;
  tp.c_oflag = OPOST | ONLCR;

  tp.c_lflag =
    ISIG | ICANON | ECHO | ECHOE | ECHOK | ECHOCTL | ECHOKE | IEXTEN;

  /* Set the settings */
  ioctl(0, TCSETA, &tp);
}

/* 
   General method for doing this comes from tinylogin getty.c 
   Copyright (C) 1999-2003 Erik Andersen <andersen@codepoet.org> 
*/

static void set_utmp(const char *tty) {

  struct utmp ut;
  struct utmp *utp;
  struct passwd *pw;

  pid_t mypid = getpid();

  utmpname(_PATH_UTMP);
  setutent();
  
  /* 
     Search by TTY id.  If it exists, then post a warning, but use it anyway.
  */

  while((utp = getutent())) {
    if (!strcmp(utp->ut_id, tty + 3)) {
      if (utp->ut_pid) {
	
	LOG("WARNING:  this utmp entry is already being used by [%d]\n", 
	    utp->ut_pid);

	utp->ut_pid = 0;
      }
      
      break;
    }
  }
    
  if (utp) 
    memcpy(&ut, utp, sizeof(ut));
  else {
    /* Start a new entry */

    memset(&ut, 0, sizeof(ut));
    strncpy(ut.ut_id, tty + 3, sizeof(ut.ut_id));
  }

  /* Get the user name */

  if ((pw = getpwuid(getuid())))
    strncpy(ut.ut_user, pw->pw_name, sizeof(ut.ut_user));
  
  strncpy(ut.ut_line, tty, sizeof(ut.ut_line));

  ut.ut_time = time(0);
  ut.ut_type = USER_PROCESS;
  ut.ut_pid = mypid;
  
  pututline(&ut);
  endutent();

  updwtmp(_PATH_WTMP, &ut);
}

/* Just like above, but clear out the entry so we don't have a messy UTMP */

static void clear_utmp(pid_t pid) {
  struct utmp ut, *utp;

  utmpname(_PATH_UTMP);
  setutent();

  while((utp = getutent())) 
    if (utp->ut_pid == pid) break;

  if (!utp) return;  /* No entry.  Oh, well */

  memcpy(&ut, utp, sizeof(ut));
  ut.ut_type = DEAD_PROCESS;
  
  ut.ut_user[0] = ut.ut_host[0] = 0;

  ut.ut_time = 0;
  ut.ut_pid = 0;
  
  pututline(&ut);
  endutent();

  updwtmp(_PATH_WTMP, &ut);
}

/* 
   Execute the specified application in the current PID.  Obviously 
   intended for child pids either from boot menu or the standard spawn app 
*/

static void exec_app(struct app *app, const char *vtname) {

  int argc = 0;
  char *args[256], *ptr = app->command;
    
  /* Parse the arguments into an argv list */
  
  while(*ptr && argc < 255) {
    char ch = 0, *e = ptr;
      
    for( ; *e && *e != ' '; e++, ch = *e);
  
    /* Replace the $VT variable with the current tty name */

    if (!strncmp(ptr, "$VT", 3)) 
      args[argc++]= (char *) vtname;
    else {
      *e = 0;
      args[argc++] = ptr; 
    }
    
    if (!ch) break;
    ptr = e + 1;
  }
    
  args[argc++] = NULL;
  
  /* Also set the VT in an environment variable so its accessable
     by our applications.  No reason to force them to read the UTMP unless they have to  
  */

  setenv("CURVT", vtname, 1);

  /* Just before we go, we finalize the terminal settings, just to make
     sure that they are sane enough.

     FIXME:  Is this step needed? 
  */

  finalize_term();

  /* Exec the application */
  execv(args[0], args);

  /* Damnation - we didn't make it.  */
  LOG("Error spawning application [%s] [%s]\n", args[0], strerror(errno));
  
  exit(-1);
}

/* Fire up the tty and show the boot menu */

static int spawn_bootmenu(int tty) {

  char vtname[7], vtfile[15];
  pid_t pid;
  
  sprintf(vtname, "tty%d", tty);
  sprintf(vtfile, "/dev/%s", vtname);
  
  LOG("Spawning bootmenu on [%s]\n", vtname);
     
  /* Fork the application to show the boot menu */
  
  if (!(pid = fork())) {
    struct app *app = 0;

    setsid();
    set_utmp(vtname);

    /* Open the TTY for stdin / stdout / stderr */

    if (open_tty(vtfile)) {
      LOG("Unable to open tty [%s]\n", vtname);
      CLOSELOG();
      exit(-1);
    }

    app = do_bootmenu(alist);
    if (app) exec_app(app, vtname); /* It shouldn't get past this */
    LOG("Unexpected error while spawning the application");
    CLOSELOG();

    exit(0);
  }
  
  return pid;
}


/* Fork and spawn the application */

static int spawn_app(struct app *app, int tty) {

  char vtname[7], vtfile[15];
  pid_t pid;
  
  sprintf(vtname, "tty%d", tty);
  sprintf(vtfile, "/dev/%s", vtname);
  
  if (!(pid = fork())) {
    setsid();
    set_utmp(vtname);

    if (open_tty(vtfile)) {
      LOG("Unable to open tty [%s]\n", vtname);
      CLOSELOG();
      exit(-1);
    }
    
    exec_app(app, vtname); 

    LOG("Unexpected error while spawning the application");
    CLOSELOG();    
    exit(0);
  }

  return pid;
}

/* Find the application that matches the name */

static struct app *find_app(struct app *list, char *name) {

  int i;

  /* If there are any numbers at the beginning of the name, skip em */

  for( ; *name && (*name >= '0' && *name <= '9'); name++);
  if (!*name) return 0;

  for(i = 0; i < MAX_APPS; i++) 
    if (list[i].name && !strcmp(list[i].name, name)) return &list[i];
  
  return 0;
}

/* Read all the files in the directory, and spawn those applications */

static int run_tty(const char *dir, struct tty *tty, struct app *list) {
  
  struct app *app = 0;   
  struct dirent **apps = 0;
  
  int n = scandir(dir, &apps, 0, versionsort);

  if (n == -1) {
    LOG("Error [%s] while scanning directory [%s]\n", 
	strerror(errno), dir);
    return -1;
  }

  if (n) {
    int i = 0;

    for(i = 0; i < n; app = 0, i++) {
      if (apps[i]->d_name[0] == '.') continue;

      /* FIXME:  Do we need to ensure that its a regular file? */

      app = find_app(list, apps[i]->d_name);
      if (!app) {
	LOG("Unable to find the application [%s]\n", apps[i]->d_name);
	continue;
      }

      if (!app->pid) {
	app->pid = spawn_app(app, tty->tty);
	
	if (app->pid > 0) {
	  tty->pid = app->pid;
	  tty->spawn++;

	  if (!tty->timestamp) tty->timestamp = time(0);

	  tty->app = app;
	  break;
	}
      }

      free(apps[i]);
    }    
  }

  /* If no applications where found, then just show the boot menu instead */

  if (!n || !app) {

    tty->pid = spawn_bootmenu(tty->tty);
    tty->spawn++;

    if (!tty->timestamp) tty->timestamp  = time(0);
    tty->app = 0;
  }

  if (apps) free(apps);

  return 0;
}

/* Loop through all the directories named ttyXX and run the applications within */
    
static int run_ttys(const char *directory, struct app *list ) {

   char ttydir[128];

   struct dirent **ttys = 0;
   int n = scandir(directory, &ttys, 0, versionsort);

   if (n == -1) {
     LOG("Error [%s] while scanning directory [%s]\n", strerror(errno), directory);
     return -1;
   }

   /* Look for all directories that start with tty */

   sprintf(ttydir, "%s/", directory);

   while(--n) {
     struct stat s;
     int tty = 0;
     
     if (strncmp(ttys[n]->d_name, "tty", 3)) continue;

     strcpy(ttydir + strlen(directory) + 1, ttys[n]->d_name);    
     if (stat(ttydir, &s) || !S_ISDIR(s.st_mode)) continue;
    
     tty = atoi(ttys[n]->d_name + 3);
     if (tty > MAX_TTYS) continue;    /* Sanity check */
     
     ttylist[tty].tty = tty;		   
     run_tty(ttydir, &ttylist[tty], list);
     
     free(ttys[0]);
   }
   
   free(ttys);
   return 0;
}

/* A special structure containing the keywords that may be present in an application file */



static struct keywords app_keywords[] = {
  KEYWORD("title:", struct app, title),
  KEYWORD("app:", struct app,  command),
  { NULL, 0 }
};

/* Read all the applications in the directory, and store them for future use */

static int read_apps(const char *directory, struct app *list, int count) {
  DIR *dir = 0;
  int apps = 0;
  char appdir[128];

  snprintf(appdir, sizeof(appdir) - 1, "%s/%s", directory, _BOOTAPPSDIR);
  appdir[127] = 0;

  if (!(dir = opendir(appdir))) {
    LOG("Unable to open directory [%s] for reading\n",
	appdir);

    return -1;
  }

  while(apps < count){
    char file[256];
    struct stat s;

    struct dirent *entry = readdir(dir);

    if (entry == NULL) break;
    if (entry->d_name[0] == '.') continue;

    snprintf(file, sizeof(file) - 1, "%s/%s", appdir, entry->d_name);
    file[255] = 0;

    if (stat(file, &s)) continue;
    if (!S_ISREG(s.st_mode)) continue;  
    
    list[apps].name = strdup(entry->d_name);

    if (!parse_file(file, app_keywords, (void *) &list[apps]))
      apps++;
  }

  closedir(dir);
  return apps;
}


static struct keywords config_keywords[] = {
  KEYWORD("logfile:", struct config, logfile),
  KEYWORD("plugin:", struct config, plugin),
  KEYWORD("plugindir:", struct config, plugindir),
  { NULL, 0 }
};

void parse_config(const char *dir) {
  char file[128];
  snprintf(file, sizeof(file) - 1, "%s/bootmenu.conf", dir);

  file[127] = 0; 

  parse_file(file, config_keywords, (void *) &bootmenu_config);
}

/* Kill everything, and bailout cleanly */
         
void kill_process(int signal) {
  
  int i;

  /* Clear out all of the processes */

  for(i = 0; i < MAX_TTYS; i++) {
    if (ttylist[i].pid) kill(ttylist[i].pid, SIGTERM);
    waitpid(ttylist[i].pid, 0, 0); /* Wait for it to die */    
    clear_utmp(ttylist[i].pid);    /* Clear out the utmp */
  }

  LOG("Now terminating. Good bye!\n");
  CLOSELOG();

  exit(0);
}
      
int main(int argc, char **argv) {

  char *dirname = _BOOTDIR;

  signal(SIGHUP, kill_process);
  signal(SIGINT, kill_process);

  while(1) {
    extern char *optarg;
    signed ch = getopt(argc, argv, "d:");

    if (ch == 'd') 
      dirname = optarg;
    else if (ch == -1) 
      break;
  }

  parse_config(dirname);
  OPENLOG();

#ifdef PLUGINS
  /* Fixme:  Technically, we can still spawn applications, even with
     no bootmenu */

  if (init_plugin(bootmenu_config.plugin)) {
    LOG("Unable to load any plugins.  I have nothing to show\n");
    CLOSELOG();
    exit(0);
  }

#endif

  /* Read the list of applications */

  if (read_apps(dirname, alist, MAX_APPS) <= 0) {
    LOG("Unable to read any applications.  I have nothing to do.\n");   
    CLOSELOG();
    exit(0); 
  }

  /* Spawn the applications for each tty in our list */
  run_ttys(dirname, alist);

  /* Now, collect the dead children and handle a respawn */
 
  while(1) {
    int i;
    pid_t pid = wait(0);
    if (pid <= 0) break;

    for(i = 0; i < MAX_TTYS; i++) {
      
      if (ttylist[i].pid == pid) {

	clear_utmp(pid);  /* Clear the utmp entry */

	if (ttylist[i].app) ttylist[i].app->pid = 0;
	ttylist[i].app = 0;

	/* Respawn unless we are respawning too quickly */

	if (!(ttylist[i].spawn % 5)) {
	  time_t now = time(0);

	  if (ttylist[i].timestamp - now < 60) {
	    ttylist[i].pid = 0;  /* No respawn */
	    LOG("Suspending tty [%d] from respawning\n", ttylist[i].tty);
	    break;
	  }

	  ttylist[i].timestamp = 0;	    
	}

	/* Show the boot menu on the recently cleared TTY */
	ttylist[i].pid = spawn_bootmenu(ttylist[i].tty);
	ttylist[i].spawn++;

	if (!ttylist[i].timestamp) ttylist[i].timestamp = time(0);	
	break;
      }
    }
  }

  LOG("NO more children - bailing out\n");
  CLOSELOG();
  return 0;
}

  

    
    
    
    

  


