/*
 * bootmenu - A program for spawning apps on a TTY
 * Developed especially for booting a graphical engine
 * but I guess it could be used for anything
 *
 * Copyright (C) 2003, Penguru Consulting, LLC
 * Written by:  Jordan Crouse (jordan@penguru.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/* Ncurses menu drawing */

/* This provides for a somewhat prettier menu, a much more portable
   interface, and some better drawing.  However, it comes at the price
   of the -ncurses library.  If we wanted to to, we could probably
   get a bit more fancy here, but for the moment, it works for me.
*/

#include <string.h>
#include <ncurses.h>
#include "bootmenu.h"

static void print_menu(WINDOW *window, int width, struct app *apps, 
		       int count, int hilite) {

  int y = 1;
  int x = 2;
  int i;

  box(window, 0, 0);
  mvwprintw(window, 0, 2, "Bootmenu");  /* Title */
 
  for(i = 0; i < count; i++) {
    if (i == hilite)
      wattron(window, A_REVERSE);

    if (strlen(apps[i].title) > width) {
      char buffer[width + 1];
      
      bzero(buffer, sizeof(buffer));
      strncpy(buffer, apps[i].title, width - 3);
      strcat(buffer, "...");
      
      mvwprintw(window, y, x, "%s", buffer);
    }
    else
      mvwprintw(window, y, x, "%s", apps[i].title);
    
    if (i == hilite)
      wattroff(window, A_REVERSE);

    y++;
  }
  
  wrefresh(window);
}
  
struct app * ncurses_bootmenu(struct app *apps) {

  WINDOW *win = 0;

  int count = 0;
  int width = 0;

  int wx, wy, ww, wh;

  int val = 0;

  for( ; apps[count].title; count++) 
    if (strlen(apps[count].title) > width) width = strlen(apps[count].title);
  
  initscr();
  clear();
  noecho();
  cbreak();

  curs_set(0);

  /* Center the window in the middle of the screen */
  /* FIXME:  temporary */

  if (width > COLS - 4) width = COLS - 4;

  ww = width + 4;
  wh = count + 2;

  wx = (COLS - ww) / 2;
  wy = (LINES - wh) / 2;
  
  win = newwin( wh, ww, wy, wx);
  keypad(win, TRUE);
    
  print_menu(win, width, apps, count, val);
  
  while(1) {
    int ch = wgetch(win);
    
    if (ch == KEY_UP) 
      val = (val == 0) ? val = count - 1 : val - 1;
    else if (ch == KEY_DOWN)
      val = (val == count - 1) ? 0 : val + 1;
    else if (ch == 10) 
      break;

    print_menu(win, width, apps, count, val);
  }

  curs_set(1);

  clrtoeol();
  refresh();
  endwin();
  
  return &apps[val];
}

/* If we are building this as a plugin, then we provide
   the standard plugin hook  

   otherwise, we provide the global do_bootmenu() function
   called by spawn_bootmenu();
*/

#ifdef PLUGINS
struct app *plugin_bootmenu(struct app *apps) 
#else
struct app *do_bootmenu(struct app *apps) 
#endif

{
  return ncurses_bootmenu(apps);
}

